/*
 * mod_antihak - Anti-HTTP-based-Worm Apache module
 * Copyright (C) 2001, Informetech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the license, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Contact:
 *  * tom@informetech.com for issues regarding mod_antihak
 *  * matt@informetech.com for issues regarding Informetech.
 *  
 */

//
// 3/10/2001 - Preparation for SourceForge.net release
// 
//   You'll have to forgive the mess - this was, after all, initially
//   a hacked out module: I hadn't written an Apache module before this :)
//   You'll notice a system() call - I'm aware that this is bad practise,
//   but as I understand exec~() calls will smash up the module in memory,
//   and we can't have that because Apache will need a return value!
//

#include <httpd.h>
#include <http_core.h>
#include <http_config.h>
#include <http_protocol.h>
#include <http_main.h>
#include <http_log.h>
#ifdef _WIN32
#include <process.h>
#else
#include <unistd.h>
#endif
#include <string.h>
#include "antihak_stats.h"

#ifdef USE_MYSQL_STATS
#include <mysql.h>
#endif

#define IPCHAINS_PATH "/sbin/ipchains"
#define SUDO_PATH "/usr/bin/sudo"

typedef struct hackentry
{
	char *name;
	char *affects;
	char *evil_uri;
	char *description;

}hackentry;

#define AFFECTS_IIS4 "IIS4.0"
#define AFFECTS_IIS5 "IIS5.0"
#define AFFECTS_SOLARIS "Sun Solaris"
#define AFFECTS_BSD "BSD"
#define AFF_AND " "
#define AFF_OR "/"

const hackentry g_vHackTable[] =
{
  { "Nimda/Hack", AFFECTS_IIS4 AFF_OR AFFECTS_IIS5, "/scripts/root.exe?", "Popular Internet Worm as-at September 2001" },
  { "CodeRed/Hack", AFFECTS_IIS4 AFF_OR AFFECTS_IIS5, "/default.ida?", "Popular Internet Worm as-at August 2001" },
  { "sadmind-iis/Hack", AFFECTS_SOLARIS AFF_OR AFFECTS_IIS4 AFF_OR AFFECTS_IIS5, "/scripts/..%c1%1c../winnt/system32/cmd.exe%20/c+dir", "Worm exploiting several old vulnerabilities" },
  { (char*)0, (char*)0, (char*)0, (char*)0 }
};

int verify_ip( const char *ip )
{
  int oct;
  int n = 0;
  char *tok, *old;

  tok = (char*)ip;
  
  while( *tok != 0 )
  {
    if( ++n > 4 )
    {
      return 0;
    }
	
    old = tok;
    tok = strchr( tok, '.' );

    if( ( tok == 0 ) )
    {
      break;
    }
	
    *tok = 0;
    oct = atoi( old );
	
    if( ( oct > 255 ) || ( oct < 0 ) )
    {
      return 0;
    }

    *tok = '.';
    tok++;
  }

  if( n != 4 )
  {
    return 0;
  }
  else
  {
    return 1;
  }
}

int verify_ips( request_rec *r )
{
  if( !verify_ip( r->connection->remote_ip ) )
  {
    return 0;
  }

  return verify_ip( r->connection->local_ip );
}

#ifdef USE_MYSQL_STATS

void antihak_add_stats( request_rec *r, hackentry *h )
{
  const char *query_fmt = "UPDATE antihak SET attempts=attempts+1 WHERE vuln=\"%s\"";
  MYSQL *conn;
  MYSQL_RES *res;
  MYSQL_ROW row;
  char *query;

  if( ( conn = mysql_init( NULL ) ) == NULL )
  {
    return;
  }

  if( mysql_real_connect( conn, "localhost", STATS_USER, STATS_PASSWD, "antihak_stats", 3306, NULL, 0 ) == NULL )
  {
    mysql_close( conn );

    return;
  }
	
  mysql_select_db( conn, "antihak_stats" );

  //
  // yes, that's a MINUS one
  //
  query = (char*)malloc( strlen( query_fmt ) + strlen( h->name ) - 1 );

  if( query == 0 )
  {
    mysql_close( conn );

    return;
  }
	
  sprintf( query, query_fmt, h->name );
  mysql_query( conn, query );
  free( query );
  mysql_close( conn );
}

#endif

int antihak_block_ip( request_rec *r )
{
    char *argv;
    const char *str_fmt = "%s %s -A input -s %s -d %s 80 -p tcp -j REJECT";
    register int fmt_len = ( strlen( r->connection->remote_ip ) +
                             strlen( r->connection->local_ip ) +
                             strlen( SUDO_PATH ) +
                             strlen( IPCHAINS_PATH ) +
                             strlen( str_fmt ) - 7 );

  if( !verify_ips( r ) )
  {
    return 0;
  }

  argv = (char*)malloc( (size_t)fmt_len );

  if( argv == 0 )
  {
    return 0;
  }

  sprintf( argv, str_fmt, SUDO_PATH, IPCHAINS_PATH, r->connection->remote_ip, r->connection->local_ip );
	
  //
  // XXXX: Find a replacement for this. execve is not good enough.
  //
  system( argv );
  free( argv );
	
  return 1;
}

int antihak_document_handler( request_rec *r )
{
  hackentry *pEntry;
  int result = DECLINED;

  ap_soft_timeout( "send", r );

  if( r->header_only )
  {
    r->content_type = "text/html";
    ap_send_http_header( r );
    ap_kill_timeout( r );

    return OK;
  }

  pEntry = (hackentry*)g_vHackTable;

  while( pEntry->name != (char*)0 )
  {
    if( strncmp( pEntry->evil_uri, r->unparsed_uri, strlen( pEntry->evil_uri ) ) == 0 )
    {
      //
      // set the content type for the outgoing header
      //
      r->content_type = "text/html";

      //
      // send the HTTP header - be nice. We're about to block 'em, after all :)
      //
      ap_send_http_header( r );
		    
      ap_rputs( DOCTYPE_HTML_3_2, r );
      ap_rputs( "<HTML><HEAD><TITLE>Attack Detected</TITLE></HEAD>", r );
      ap_rputs( "<BODY><P>", r );
      ap_rprintf( r, "<H2>You (%s) are banned</H2></P>", r->connection->remote_ip, r->connection->local_host );
      ap_rprintf( r, "<P><I>Detected: %s<BR>Affects: %s<BR>Description: %s<BR>Contact: %s</I></P>", pEntry->name, pEntry->affects, pEntry->description, r->server->server_admin );
      ap_rprintf( r, "<HR><P><FONT size=\"1\">mod_antihak, (c) 2001, by Tom Lee (tlee1@iprimus.com.au)</FONT></P></BODY></HTML>" );
			
      ap_log_printf( r->server, "[mod_antihak]: Detected %s attack from %s. Blocking IP ...", pEntry->name, r->connection->remote_ip );

      if( !antihak_block_ip( r ) )
      {
        ap_log_printf( r->server, "[mod_antihak]: Failed to allocate memory. IP not blocked." );
      }
			
#ifdef USE_MYSQL_STATS
      antihak_add_stats( r, pEntry );
#endif
      result = OK;

      break;
    }

    pEntry++;
  }

  ap_kill_timeout( r );

  return result;
}

handler_rec antihak_handlers[] =
{
  { "*", antihak_document_handler },
  { NULL }
};

module MODULE_VAR_EXPORT antihak_module =
{
	STANDARD_MODULE_STUFF,
	NULL, 
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	antihak_handlers,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

